' Todo - code for Read from LCD
' Todo - allow 8 data lines
' Todo - split to 3 (?) documents.
' Todo - sub LCDHelp to display all subs and parameters
' Todo - establish naming conventions for parameters, subs
' Todo - doco.  Include reference to tutorials.
' Todo - test with all types of LCD
' Todo - add test to check version >= 3.1
' Todo - write LCD_Display, LCD_Cursor, LCD_Blink (ON/OFF)
' Todo - test for operations thst will fail for large fonts
' ToDo - Pin RW set as open collector.  What should it be?



'
' demonstration program
' send lines to the LCD
'
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' LCD driver for MMBasic 3.1 or later (uses defined subroutines)
' Geoff Graham, Jan 2012.
' Modified by James Deakins to include 20 x 4 LCDs, Feb 2012.
'
' Will drive standard 16 x 2 or 20 x 4 LCDs
' For example:  futurlec.com       LCD16X2
'               altronics.com.au   Z7001
'               jaycar.com.au      QP5512
'                                  HD44780 (20 x 4)
'
' To use:
'- Optionally, redefine the pins used with the command: ChangeLCDpins RS, EN, D1, D2, D3, D4, RW
'- Define LCD size with the command:  InitLCDSize columns, lines
'- Setup the LCD with the command:    InitLCD
'- Display a line using the command:  PrintLCD LineNbr, Text$
'
' See documentation (LCD.PDF) for the schematic
' Maximite pin 11 is RS, pin 12 is EN
' pins 13 to 16 are D4 to D7
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' following should go to the initialisation file, for insertion at the beginning of code
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

' The following block of variables should not change once their values are set
DIM C_LineStart(4)     ' an array to hold the start addresses for each line
C_Lines = 0            ' variable to hold the number of lines on the LCD
C_LineLength = 0       ' variable to hold the line length of the LCD
'
' The following block of variables are for internal use by the subroutines,
' mainly to supply missing information when the user only changes one attribute.
' For instance, to turn Blink on, you normally have to supply 2 other parameters.

' We STRONGLY suggest that users only use subroutines to change these variables.
' You could change them manually, but there's no guarantee they will work and, importantly,
' your code may fail with future versions of the library as more functionality is added to
' the subs.

_LCD_Display$   = ""
_LCD_Cursor$    = ""
_LCD_Blink$     = ""
_LCD_Type$      = ""
_LCD_MoveDir$   = ""
_LCD_Shift$     = ""
_LCD_Type$      = ""
_LCD_Direction$ = ""
_LCD_Bits       = 0
_LCD_Lines      = 0
_LCD_Font$      = ""
_LCD_RWenabled  = 0 ' variable indicating whether the Read Write pin is enabled (1=True)

' Allocate names to pin numbers based on LCD.PDF and the comments above.
' The pin numbers must be changed using the ChangeLCDPins sub.

RS = 11
EN = 12
D1 = 13
D2 = 14
D3 = 15
D4 = 16
'
'
' This is where to put the subroutine calls to set up the library and LCD hardware.

' **** Change the parameters to show whether you have a 16 x 2, or a 20 x 4 LCD screen.
InitLCDSize(20, 4)

' if you need to change pin numbers, do it before calling InitLCD
' eg ChangeLCDPins(5, 6, 7, 8, 9, 10[,11])   ' - don't use unless you need to

InitLCD ' will need parameter for 4 or 8 bit

' Following should go to demo/doco file
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

' display some simple lines, one at a time

'PrintLCD 1, "  Hello  World"
'PrintLCD 2, "  Maximite LCD"
'PrintLCD 3, "with a 20 x 4 LCD,"
'PrintLCD 4, "using Subroutines."

TestSubs

'--------------------------------------

SUB TestSubs
' Code to test subs
'    TestLCDClearScreen ok
'    TestLCDHome ok
    TestLCDEntryMode
'    TestLCDControl ok
'    TestLCDShift ok
'    TestLCDFunctionSet
'    TestLCDSetCGRAMaddr
'    TestLCDSetDDRAMaddr
'    TestLCDreadBusyFlag
    LCDClearScreen
    PrintLCD(1, "!!END OF TESTS!!")

End Sub


SUB TestLCDClearScreen

    PrintLCD(1,"Screen Clearing")
    PrintLCD(2,"in 2 seconds")
    Pause 2000
    LCDClearScreen
    Pause 2000
    PrintLCD(1,"Screen should")
    PrintLCD(2,"have cleared")
    ? "Please record whether the screen cleared."
    WaitForUser
    LCDClearScreen

END SUB


SUB TestLCDHome

    LCD_Cursor("OFF")
    LCD_Blink("ON")
    PrintLCD(1,"Cursor should")
    PrintLCD(2,"move to Home")
    Pause 2000
    LCDHome
    ? "Cursor is OFF, Blink is ON, IMHO, blinking insert point should be at Home position"
    WaitForUser
    LCDClearScreen

END SUB


SUB TestLCDEntryMode
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Sets cursor move direction and specifies display shift.
' These operations are performed during data write and read.
' argument #1 is the move direction ("I" (Increment) or "D" (Decrement))
' argument #2 is the display shift ("On" or "Off")
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDEntryMode(MoveDir$, Shift$)
	LOCAL StartCol, i, Msg$
	
	StartCol = C_LineLength \ 2		' start half way across the display
    LCDClearScreen
    PrintLCD(1,"Entry Mode Test")
	LCDLocate(2,StartCol)
	? "Cursor Move direction = Increment, Display Shift = ON"
	? "Display shifts to the Left as the message on line 2 appears.
	? "Message is to the LEFT of column StartCol."
    LCDEntryMode("I", "On")
    Msg$ = "A test message"
    FOR i = 1 TO LEN(Msg$)
        LCD_WriteByte ASC(MID$(Msg$, i, 1)), 1
        PAUSE 500
    NEXT i
	WaitForUser

    LCDClearScreen
    LCDEntryMode("I", "Off")
    PrintLCD(1,"Entry Mode Test")
	LCDLocate(2,StartCol)
	? "Cursor Move direction = Increment, Display Shift = OFF"
	? "Display stays still as the message on line 2 appears.
	? "Message is to the RIGHT of column StartCol."
    Msg$ = "A test message"
    FOR i = 1 TO LEN(Msg$)
        LCD_WriteByte ASC(MID$(Msg$, i, 1)), 1
        PAUSE 500
    NEXT i
	WaitForUser	    

    LCDClearScreen
    LCDEntryMode("D", "On")
    PrintLCD(1,"Entry Mode Test")
	LCDLocate(2,StartCol)
	? "Cursor Move direction = Decrement, Display Shift = ON"
	? "Display shifts to the right as the message on line 2 appears.
	? "Message is to the LEFT of column StartCol."
    Msg$ = "A test message"
    FOR i = 1 TO LEN(Msg$)
        LCD_WriteByte ASC(MID$(Msg$, i, 1)), 1
        PAUSE 500
    NEXT i
	WaitForUser		

    LCDClearScreen
    LCDEntryMode("D", "Off")
    PrintLCD(1,"Entry Mode Test")
	LCDLocate(2,StartCol)
    Msg$ = "A test message"
    FOR i = 1 TO LEN(Msg$)
        LCD_WriteByte ASC(MID$(Msg$, i, 1)), 1
        PAUSE 500
    NEXT i
	WaitForUser			

END SUB


SUB TestLCDControl
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set LCD display, cursor, and cursor blink controls
' argument #1 is the Display      status ("On" or "Off")
' argument #2 is the Cursor       status ("On" or "Off")
' argument #3 is the Cursor Blink status ("On" or "Off")
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDControl(Display$, Cursor$, Blink$)

    PrintLCD 1, "  Hello  World"
    IF C_Lines > 1 THEN PrintLCD 2, "  Maximite LCD"
    IF C_Lines > 2 THEN PrintLCD 3, "with a 20 x 4 LCD,"
    IF C_Lines > 3 THEN PrintLCD 4, "using Subroutines."
    PAUSE 2000

    ' Tester should see the screen clear, and a blinking cursor
    '
    PrintLCD(1, "CLR SCR,Disp on")
    PrintLCD(2,"Cur on,Cur blink")
    ? "The screen should clear, the cursor should be on, and blinking"
    

    lcdClearScreen
    PrintLCD(1, "Screen clear")
    WaitForUser

    LCDControl("On", "On", "On")    ' Display on, Cursor on, Cursor blink on
    PrintLCD(1, "Cur Blink ON")
    WaitForUser

    LCDControl("On", "On", "Off")
    PrintLCD(2, "Blink OFF")
    ? "The Cursor should be ON, but not blinking"
    WaitForUser
    lcdClearScreen

    LCDControl("On", "Off", "On")
    PrintLCD(1, "Cur OFF,Blink ON")
    ? "The Cursor should be OFF, and Blinking ON"
    WaitForUser
    ' lcdClearScreen

    LCDControl("On", "Off", "Off")
    PrintLCD(2, "Cur OFF,Blnk OFF")
    ? "The Cursor should be OFF, and not blinking"
    WaitForUser
    lcdClearScreen

    PrintLCD(1, "Display will OFF")
    Pause 2000
    LCDControl("Off", "On", "On")
    ? "The Display should be OFF, with a blinking Cursor (not visible, as display is off)"
    WaitForUser

    LCDControl("On", "On", "On")
    PrintLCD(2, "Display back ON")
    ? "The Display is back ON, with a blinking cursor"
    WaitForUser
    lcdClearScreen

END SUB


SUB TestLCDShift

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Cursor and Display shift.
' Moves cursor and shifts display without changing DDRAM contents.
' argument #1 is the type ("S" (display Shift) or "C" (Cursor move)
' argument #2 is the shift direction ("L" (Left) or "R" (Right))
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDShift(Type$, Direction$)

    Local TestMsg$

    PrintLCD(1, "Cursor & Display")

' Load a 26 char message into Line 2 (noting that on a 16 x 2 character display
' we only see the first 16 characters, a 4 line display will wrap the text to
' line 4.)
' ------------------------------------------------------------------------------
    TestMsg$ = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    PrintLCD(2, TestMsg$)
    Pause 2000
'
' Now shift the message so the balance of the second line can be read
' -------------------------------------------------------------------
'
    ? "All lines will be Shifted Left for 40 characters"
    WaitForUser
    LCDLocate(2,1)
    for _LCD = 1 to 40
        LCDShift("S", "L")
        pause 500
    next _LCD
    ? "End of test for Display Shift Left. Cursor should be where it started."
    WaitForUser
'
' Note that this shifts both lines left, and you have to shift left 40 times
' to get back where you started from
'-------------------------------------------------------------------------------------
' Also notice how the text may change to different lines as the message is scrolled.
'

    ? "All lines will be Shifted Right for 40 characters"
    WaitForUser
    LCDLocate(2,1)
    for _LCD = 1 to 40
        LCDShift("S", "R")
        pause 500
    next _LCD
    ? "End of test for Display Shift Right. Cursor should be where it started."
    WaitForUser

    ? "Cursor will be Shifted Left for 39 characters.  Display won't move."
    WaitForUser
    LCDLocate(2,40)	' end of line 2 - on 4 line display will be end of line 4
    for _LCD = 1 to 39 ' 39 not 40, because 'A' already in posn 1. 
        LCDShift("C", "L")
        pause 500
    next _LCD
    ? "End of test for Cursor Shift Left. Cursor should be where it started."
    WaitForUser

    ? "Cursor will be Shifted Right for 39 characters.  Display won't move."
    WaitForUser
    LCDLocate(2,1)
    for _LCD = 1 to 39
        LCDShift("C", "R")
        pause 500
    next _LCD
    ? "End of test for Cursor Shift Right. Cursor should be at end of line."
    WaitForUser

END SUB


SUB TestLCDFunctionSet
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the interface data length, number of display lines and
' character font.
' argument #1 is the number of bits used by the interface (4 or 8)
' note: this library is currently only set up to use 4 bits
' argument #2 is the number of lines (1 or 2)
' TODO: Update with how to select 4 lines
' argument #3 is the font size ("L" (Large) or "N" (Normal))
' Large is 5 x 10 dots, Normal is 5 x 8 dots.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDFunctionSet(_Bits, _Lines, _Font$)

LCDFunctionSet(4, 1, "N")
LCDFunctionSet(4, 1, "L")
LCDFunctionSet(4, 2, "N")
LCDFunctionSet(4, 2, "L")
LCDFunctionSet(8, 1, "N")
LCDFunctionSet(8, 1, "L")
LCDFunctionSet(8, 2, "N")
LCDFunctionSet(8, 2, "L")

END SUB


SUB TestLCDSetCGRAMaddr
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Character Generator RAM (CGRAM) address.
' CGRAM data is sent and received after this setting.
' argument #1 is the CGRAM address (6 bits, so 0 to 63)
'
' In the character generator RAM, the user can rewrite character
' patterns by program.
' For 5 x  8 dots, eight character patterns can be written, and
' for 5 x 10 dots, four  character patterns can be written.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDSetCGRAMaddr(_Addr)
END SUB


SUB TestLCDSetDDRAMaddr
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Character Generator RAM (CGRAM) address.
' DDRAM data is sent and received after this setting.
' argument #1 is the DDRAM address (7 bits, so 0 to 127)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDSetDDRAMaddr(_Addr)
END SUB


SUB TestLCDreadBusyFlag
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Read busy flag.
' Returns the Busy Flag (BF) indicating internal operation is being
' performed and returns address counter contents.
' bf=1 - still internally operating. b=0 - instructions acceptable
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDreadBusyFlag(_Status, _Addr)
END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Wait for the user to press any key on the keyboard
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB WaitForUser
	LOCAL Key$
	
	? "Press any key to continue " ;
	Do
		Key$ = INKEY$
	LOOP UNTIL Key$ <> ""
	?
	
END SUB

' and here's a test of one of the subroutines...
' TestMsg$ = "This is a really long message that we will scroll across the second line of the LCD screen."
' ScrollLCDMessage(2, TestMsg$)


' remainder (all subs) should go to the .lib file
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Initialise the size based variables for this application
' (C_Prefix indicates constants).
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB InitLCDSize(LCD_LineLength, LCD_Lines)

    C_LineLength = LCD_LineLength
    C_Lines = LCD_Lines
    '
    C_LineStart(1) = 0        ' Start address for line 1
    C_LineStart(2) = 64       ' Start address for line 2
    C_LineStart(3) = 20       ' Start address for line 3
    C_LineStart(4) = 84       ' Start address for line 4

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Change the Maximite / DM pins allocate for LCD control and data.
' argument #1 is the pin to use for the LCD RS line
' argument #2 is the pin to use for the LCD EN line
' argument #3 is the pin to use for the LCD D1 line
' argument #4 is the pin to use for the LCD D2 line
' argument #5 is the pin to use for the LCD D3 line
' argument #6 is the pin to use for the LCD D4 line
' argument #7 is the pin to use for the LCD RW line (optional)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB ChangeLCDPins(_RS, _EN, _D1, _D2, _D3, _D4, _RW)
    RS = _RS
    EN = _EN
    D1 = _D1
    D2 = _D2
    D3 = _D3
    D4 = _D4
    IF _RW <> 0 Then
        IF RWenabled = 1 THEN   ' true
            RW = _RW
        ELSE
            ? "You must allow use of the Read/Write line by calling sub LCDenableRW"
        ENDIF
    ENDIF
END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Initialise the LCD hardware
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB InitLCD
    SETPIN RS, 9                                ' open collector
    SETPIN EN, 9                                ' open collector
    SETPIN D1, 9                                ' open collector
    SETPIN D2, 9                                ' open collector
    SETPIN D3, 9                                ' open collector
    SETPIN D4, 9                                ' open collector
    IF RWenabled THEN SETPIN RW, 9              ' open collector (optional)

    LCD_Nibble_Write &B0011, 0, 5                    ' reset
    LCD_Nibble_Write &B0011, 0, 5                    ' reset
    LCD_Nibble_Write &B0011, 0, 5                    ' reset
    LCD_Nibble_Write &B0010, 0, 2                    ' 4 bit mode
    LCD_Nibble_Write &B0010 : LCD_Nibble_Write &B1100       ' 4 bits, 2 lines
    LCD_Nibble_Write &B0000 : LCD_Nibble_Write &B1100       ' display on, no cursor
    LCD_Nibble_Write &B0000 : LCD_Nibble_Write &B0110       ' increment on write
    LCD_Nibble_Write &B0000 : LCD_Nibble_Write &B0001, 0, 2 ' clear the display, pause 2
END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Display a line on the LCD
' argument #1 is the line to be used (1 to maximum line number)
' argument #2 is the line of characters to be displayed (any length)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB PrintLCD ( LineNumber, Line$ )
    LOCAL i, c, fillSpaces, fullLine$, controlByte

    controlByte = 128 + C_LineStart(LineNumber)

    ' first, send the cursor position (in two nibbles)
    ' LCD_Nibble_Write INT(controlByte/16) : LCD_Nibble_Write controlByte
    LCD_WriteByte(controlByte)

    ' next, make sure the line is right padded with spaces, if needed
    fillSpaces = C_LineLength - LEN(Line$)
    IF fillSpaces > 0 THEN
        fullLine$ = Line$ + SPACE$(fillSpaces)
    ELSE
        fullLine$ = Line$
    ENDIF

    ' then send the text character by character (two nibbles per character)
    ' note - different to LCD.BAS. This allows text longer than line length.
    FOR i = 1 TO LEN(Line$)
        ' c = ASC(MID$(fullLine$, i, 1))
        ' LCD_Nibble_Write INT(c/16), 1 : LCD_Nibble_Write c, 1
        LCD_WriteByte ASC(MID$(fullLine$, i, 1)), 1
    NEXT i

END SUB


SUB LCD_WriteByte(_Byte, _Flag, _Wait_mSec)

    LCD_Nibble_Write INT(_Byte/16), _Flag, _Wait_mSec
    LCD_Nibble_Write _Byte, _Flag, _Wait_mSec

END SUB


SUB LCD_ReadByte(_Byte)
    LOCAL myByte, b

    LCD_Nibble_Read(b)
    myByte = b  * 16
    LCD_Nibble_Read(b)
    myByte = myByte OR b
    ? myByte; " "; BIN$(myByte)

    _Byte = myByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Read the lower 4 bits (called a nibble) from the LCD
' argument #1 returns the nibble
' argument #2 is delay after the data has been received (default is zero)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCD_Nibble_Read ( Data, Wait_mSec )
    LOCAL Value
    ' still under development
    PIN(RS) = 1                 ' data
    PIN(EN) = 1 : PIN(EN) = 0
    IF Wait_mSec > 0 THEN PAUSE Wait_mSec       ' modified 30/1/2012. Restore to Wait_mSec when Pause 0 bug corrected.
    ? "Bits read are: ; PIN(D1); " "; PIN(D2); " "; PIN(D3); " "; PIN(D4)
    Value = PIN(D1) ' low order bit
    Value = Value OR (PIN(D2) * 2)
    Value = Value OR (PIN(D3) * 4)
    Value = Value OR (PIN(D4) * 8)

    ? "Nibble value is "; Value; " "; Bin$(Value)

    Data = Value

END SUB
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Send the lower 4 bits (called a nibble) to the LCD
' argument #1 is the nibble to send
' argument #2 is true if data, false if command (default is command)
' argument #3 is delay after the data has been sent (default is zero)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCD_Nibble_Write ( Data, Flag, Wait_mSec )
    PIN(RS) = Flag
    PIN(D1) = Data AND &B00000001
    PIN(D2) = Data AND &B00000010
    PIN(D3) = Data AND &B00000100
    PIN(D4) = Data AND &B00001000
    PIN(EN) = 1 : PIN(EN) = 0
    IF Wait_mSec > 0 THEN PAUSE Wait_mSec       ' modified 30/1/2012. Restore to Wait_mSec when Pause 0 bug corrected.
END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Move cursor to a location on the LCD screen.
' argument #1 is the line to be used (1 to maximum line number)
' argument #2 is the column to be used (1 to line length)
' tested OK.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDLocate(Line, Column)
    LOCAL controlByte

    controlByte = 128 + C_LineStart(Line) + Column - 1
    LCD_WriteByte controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Scroll a message across the LCD screen
' argument #1 is the line to be used (1 to 4)
' argument #2 is the message to display
' tested ok.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

SUB ScrollLCDMessage(line, Message$)
    LOCAL MsgLen, Mask_Offset, i, c

    MsgLen = LEN(Message$)
    FOR Mask_Offset = 0 TO MsgLen - C_LineLength ' keep track of where we're starting the display window
        LCDLocate(line,1)                     ' go to the start of line
        FOR i = 1 TO C_LineLength
            c = ASC(MID$(Message$, Mask_Offset + i, 1))
            LCD_WriteByte c, 1
        NEXT i
        IF Mask_offset = 0 THEN
            PAUSE 1000  ' initial - so user has time to read the initial part of the message
        ELSE
            PAUSE 400   ' each char - so screen has time to refresh, and user can read it
        ENDIF
    NEXT Mask_Offset

END SUB


' *** Example - Print a non-standard char. from the char. set held on the LCD
' ---------------------------------------------------------------------------
'_LCD_byte = &HF4 : pin(11) = 1 : SendByte ' print the ohm symbol
' ' note that the cursor didn't need to be moved
pause 2000 ' time to observe result

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Enable the Read Write Flag.
' Most people won't use this, but it's here for completeness and
' so the Busy Flag and address can be read from the LCD
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDenableRW

    RWenabled = 1   ' set to true

END SUB

' Set LCD display, cursor, and cursor blink controls

SUB LCD_Display(Param$)    ' "On" or "Off"

    LCDControl(Ucase$(Param$), _LCD_Cursor$, _LCD_Blink$)

END SUB

SUB LCD_Cursor(Param$)     ' "On" or "Off"

    LCDControl(_LCD_Display$, Ucase$(Param$), _LCD_Blink$)

END SUB

SUB LCD_Blink(Param$)      ' "On" or "Off"

    LCDControl(_LCD_Display$, _LCD_Cursor$, Ucase$(Param$))

END SUB


'****************************************************************
' The following subroutines implement the LCD controls.
'****************************************************************

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Clear the LCD screen.
' Clears entire display and sets DDRAM address 0 in address counter
' tested ok
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDClearScreen
    LOCAL controlByte

    controlByte = &B00000001
    LCD_WriteByte controlByte

END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set cursor to Home position.
' Sets DDRAM address 0 in address counter.  Also returns display
' from being shifted to original position. DDRAM contents remain
' unchanged.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDHome
    LOCAL controlByte

    controlByte = &B00000010
    LCD_WriteByte controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Sets cursor move direction and specifies display shift.
' These operations are performed during data write and read.
' argument #1 is the move direction ("I" (Increment) or "D" (Decrement))
' argument #2 is the display shift ("On" or "Off")
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDEntryMode(MoveDir$, Shift$)
    LOCAL controlByte

    controlByte = &B00000100

    _LCD_MoveDir$ = UCASE$(MoveDir$)
    IF _LCD_MoveDir$ = "I" THEN
        controlByte = controlByte OR &B00000010
    ENDIF

    _LCD_Shift$ = UCASE$(Shift$)
    IF _LCD_Shift$ = "ON" THEN
        controlByte = controlByte OR &B00000001
    ENDIF

    ' ? Bin$(controlByte)

    LCD_WriteByte controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set LCD display, cursor, and cursor blink controls
' argument #1 is the Display      status ("On" or "Off")
' argument #2 is the Cursor       status ("On" or "Off")
' argument #3 is the Cursor Blink status ("On" or "Off")
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDControl(Display$, Cursor$, Blink$)
    LOCAL controlByte

    controlByte = &B00001000

    _LCD_Display$ = UCASE$(Display$)
    IF _LCD_Display$ = "ON" THEN
        controlByte = controlByte OR &B00000100
    ENDIF

    _LCD_Cursor$ = UCASE$(Cursor$)
    IF _LCD_Cursor$ = "ON" THEN
        controlByte = controlByte OR &B00000010
    ENDIF

    _LCD_Blink$ = UCASE$(Blink$)
    IF _LCD_Blink$ = "ON" THEN
        controlByte = controlByte OR &B00000001
    ENDIF

    LCD_WriteByte controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Cursor and Display shift.
' Moves cursor and shifts display without changing DDRAM contents.
' argument #1 is the type ("S" (display Shift) or "C" (Cursor move)
' argument #2 is the shift direction ("L" (Left) or "R" (Right))
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDShift(Type$, Direction$)
    LOCAL controlByte

    controlByte = &B00010000

    _LCD_Type$ = UCASE$(Type$)
    IF _LCD_Type$ = "S" THEN
        controlByte = controlByte OR &B00001000
    ENDIF

    _LCD_Direction$ = UCASE$(Direction$)
    IF _LCD_Direction$ = "R" THEN
        controlByte = controlByte OR &B00000100
    ENDIF

    ' ? Bin$(controlByte)

     LCD_WriteByte controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the interface data length, number of display lines and
' character font.
' argument #1 is the number of bits used by the interface (4 or 8)
' note: this  code is currently only set up to use 4 bits
' argument #2 is the number of lines (1 or 2)
' TODO: Update with how to select 4 lines
' argument #3 is the font size ("L" (Large) or "N" (Normal))
' Large is 5 x 10 dots, Normal is 5 x 8 dots.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDFunctionSet(_Bits, _Lines, _Font$)
    LOCAL controlByte

    controlByte = &B00100000

    _LCD_Bits = _Bits
    IF _Bits = 8 THEN
        ? "Warning: 8 bit mode not currently supported!"
        ' controlByte = controlByte OR &B00010000
        EXIT SUB
    ENDIF

    _LCD_Lines = _Lines
    IF _Lines = 2 THEN   ' TODO confirm 4 lines uses same control bit
        controlByte = controlByte OR &B00001000
    ENDIF

    _LCD_Font$ = UCASE$(_Font$)
    IF _LCD_Font$ = "L" THEN
        controlByte = controlByte OR &B00000100
    ENDIF

    ' ? Bin$(controlByte)

    LCD_WriteByte controlByte

END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Character Generator RAM (CGRAM) address.
' CGRAM data is sent and received after this setting.
' argument #1 is the CGRAM address (6 bits, so 0 to 63)
'
' In the character generator RAM, the user can rewrite character
' patterns by program.
' For 5 x  8 dots, eight character patterns can be written, and
' for 5 x 10 dots, four  character patterns can be written.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDSetCGRAMaddr(_Addr)
    LOCAL controlByte

    controlByte = &B01000000

    IF _Addr > 63 THEN
        ? "CGRAM Address "; _Addr; " is too high (it should be 0 to 63)"
    ELSE
        controlByte = controlByte OR _Addr
        ? Bin$(controlByte)
        LCD_WriteByte controlByte
    ENDIF

END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Character Generator RAM (CGRAM) address.
' DDRAM data is sent and received after this setting.
' argument #1 is the DDRAM address (7 bits, so 0 to 127)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDSetDDRAMaddr(_Addr)
    LOCAL controlByte

    controlByte = &B10000000

    IF _Addr > 127 THEN
        ? "DDRAM Address "; _Addr; " is too high (should be 0 to 127)"
    ELSE
        controlByte = controlByte OR _Addr
        ? Bin$(controlByte)
        LCD_WriteByte controlByte
    ENDIF

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Read busy flag.
' Returns the Busy Flag (BF) indicating internal operation is being
' performed and returns address counter contents.
' bf=1 - still internally operating. b=0 - instructions acceptable
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDreadBusyFlag(_Status, _Addr)
    LOCAL myByte, BF, AC

    LCD_ReadByte(myByte)
    BF = myByte AND &B10000000  ' busy flag
    AC = myByte AND &B01111111  ' Address Counter

    _Status = (BF = 128)    'convert to a value of 0 or 1

END SUB

